// partition:  a container for organizing a set of sequential integers
//             into non-overlapping groups.
// Version:    0.1
// Copyright 2013 Matthew T. Busche.
//
// This program is free software:  you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
// more details.
//
// You should have received a copy of the GNU General Public License along
// with this program.  If not, see <http://www.gnu.org/licenses/>.

package org.mattbusche.util.partition;

/** <p>A bidirectional iterator for traversing the {@link Entry} objects of a
 ** {@link Group}.</p>
 **
 ** <p><code>GroupIterator</code> design differs significantly from the design
 ** of iterators in the Java Collections Framework (JCF):</p>
 ** <ol>
 ** <li><code>GroupIterator</code>s point to a particular list
 ** <code>Entry</code>.  In contrast, iterators from the JCF point
 ** between two list entries.</li>
 ** <li><code>GroupIterator</code>s are only invalidated by those operations
 ** that remove the <code>Entry</code> they point to from their
 ** <code>Group</code>.  In particular, memory reallocation events do not
 ** invalidate <code>GroupIterator</code>s.  In contrast, any update to a JCF
 ** container necessarily invalidates all iterators on that container except
 ** the one iterator used to perform that update (if any).</li>
 ** <li>The behavior of an invalidated <code>GroupIterator</code> is
 ** undefined.  In contrast, any use of an invalidated iterator from the JCF,
 ** results in the well-defined behavior of throwing a
 ** <code>ConcurrentModificationException</code>.</li>
 ** </ol>
 **
 ** <p>Because <code>GroupIterator</code>s point to a particular list entry
 ** and because they are not regularly invalidated, they can be used as
 ** semi-permanent references to Group entries.  This is at the cost of user
 ** safety: accidental use of an invalidated iterator will not produce an
 ** exception, but instead has undefined behavior.</p>
 **
 ** <p><i>Notes on GroupIterator invalidation:</i></p>
 **
 ** <p>The behavior of an invalidated <code>GroupIterator</code> is undefined
 ** so one must take care to avoid unintentional iterator invalidation.  For
 ** this Java-based implementation of partition, the only way a
 ** <code>GroupIterator</code> can become invalidated is for the
 ** <code>Entry</code> pointed to by the <code>GroupIterator</code> to be
 ** removed from its <code>Group</code>.  This may seem obvious, but remember
 ** that any operation that inserts an <code>Entry</code> to a
 ** <code>Group</code> also removes that same <code>Entry</code> from its
 ** previous <code>Group</code>; so <i>all</i> operations that affect
 ** <code>Group</code> membership can invalidate <code>GroupIterator</code>s.
 ** Also note that the <code>Domain</code>'s resize methods can be used to
 ** eliminate <code>Entry</code>s from the <code>Domain</code> entirely (and
 ** from their <code>Group</code>s).</p>
 **/
public class GroupIterator<M,V>
{
    Group<M,V> _group;
    Node<M,V> _node;

    /** Copy constructor.
     **/
    public GroupIterator(GroupIterator<M,V> iter)
    {
        this._group = iter._group;
        this._node = iter._node;
    }

    GroupIterator(Group<M,V> group, Node<M,V> node)
    {
        _group = group;
        _node = node;
    }

    /** Increments the iterator to the next list position and returns a
     ** reference to itself.
     **/
    public GroupIterator<M,V> inc()
    {
        _node = _node._next;
        return this;
    }

    /** Decrements the iterator to the previous list position and returns a
     ** reference to itself.
     **/
    public GroupIterator<M,V> dec()
    {
        _node = _node._prev;
        return this;
    }

    /** Returns <code>true</code> if this iterator is pointing one logical
     ** position before the first {@link Entry} in the list, and
     ** <code>false</code> if it is pointing to any list {@link Entry}.
     **/
    public boolean isBeforeFront()
    {
        return _node == _group._end;
    }

    /** Returns <code>true</code> if this iterator is pointing one logical
     ** position after the last {@link Entry} in the list,  and
     ** <code>false</code> if it is pointing to any list <code>Entry</code>.
     **/
    public boolean isAfterBack()
    {
        return _node == _group._end;
    }

    /** Returns <code>true</code> if this iterator points to the first
     ** <code>Entry</code> of a non-empty list, or to the afterBack position
     ** of an empty list; and <code>false</code> otherwise.
     **/
    public boolean isAtFront()
    {
        return _node == _group._end._next;
    }

    /** Returns <code>true</code> if this iterator points to the last {@link
     ** Entry} of a non-empty list, or to the beforeFront position of an empty
     ** list; and <code>false</code> otherwise.
     **/
    public boolean isAtBack()
    {
        return _node == _group._end._prev;
    }

    /** Returns the {@link Entry} object pointed to by this iterator.
     **
     ** @throws PartitionException if this iterator {@link #isBeforeFront} or
     ** {@link #isAfterBack}.
     **/
    public Entry<M> getEntry()
    {
        if(_node == _group._end)
            throw new PartitionException("Iterator is before front or after back.");
        return _node._entry;
    }

    /** Returns the {@link Group} associated with this iterator.
     **/
    public Group<M,V> getContainer()
    {
        return _group;
    }

    /** <p>Removes the {@link Entry} with the given <code>id</code> from the
     ** {@link Group} with which it was previously a member (if any), and adds
     ** it into this iterator's <code>Group</code> immediately before the
     ** position pointed to by this iterator.</p>
     **
     ** <p>Computational Complexity:  Constant.</p>
     **
     ** @throws NullPointerException if this iterator's <code>Group</code> is
     ** not associated with a <code>Domain</code>.
     ** @throws ArrayIndexOutOfBoundException if <code>id</code> is not a
     ** member of the associated Domain.
     **/
    public void addBefore(int id)
    {
        _group._insertAfter(_node._prev, id);
    }

    /** <p>Removes the {@link Entry} with the given <code>id</code> from the
     ** {@link Group} with which it was previously a member (if any), and adds
     ** it into this iterator's <code>Group</code> immediately after the
     ** position pointed to by this iterator.</p>
     **
     ** <p>Computational Complexity:  Constant.</p>
     **
     ** @throws NullPointerException if this iterator's <code>Group</code> is
     ** not associated with a <code>Domain</code>.
     ** @throws ArrayIndexOutOfBoundException if <code>id</code> is not a
     ** member of the associated <code>Domain</code>.
     **/
    public void addAfter(int id)
    {
        _group._insertAfter(_node, id);
    }

    /** <p>Removes and returns the {@link Entry} pointed to by this iterator
     ** from its {@link Group}'s list, and updates this iterator to point to
     ** the next <code>Entry</code> in the list.</p>
     **
     ** <p>Computational Complexity:  Constant.</p>
     **
     ** @throws PartitionException if this iterator {@link #isBeforeFront} or
     ** {@link #isAfterBack}.
     **/
    public Entry<M> removeAndInc()
    {
        if(_node == _group._end)
            throw new PartitionException("Iterator is before front or after back.");
        Node<M,V> n = _node;
        inc();
        _group._remove(n);
        return n._entry;
    }


    /** <p>Removes and returns the {@link Entry} pointed to by this iterator
     ** from its {@link Group}'s list, and updates this iterator to point to
     ** the previous <code>Entry</code> in the list.</p>
     **
     ** <p>Computational Complexity:  Constant.</p>
     **
     ** @throws PartitionException if this iterator {@link #isBeforeFront} or
     ** {@link #isAfterBack}.
     **/
    public Entry<M> removeAndDec()
    {
        if(_node == _group._end)
            throw new PartitionException("Iterator is before front or after back.");
        Node<M,V> n = _node;
        dec();
        _group._remove(n);
        return n._entry;
    }
}
